/********************************************************
  
  RFID-enabled light level control 
  
  Author: Jason Webb
  Website: http://cs.unk.edu/~webb/
  
  NORMAL MODE --------------------------------------------
  - System will keep an eye out for RFId tag data. When
    tag data is received from RFID reader we will read
    in entire tag, verify it, then load the color values
    associated with tag.
    
  PROGRAMMING MODE ----------------------------------------
  - System only goes into programming mode when module is
    attached.
  - System will look for RFID tag data. When data is
    received the module is then polled until user hits
    the "Set" button on the module, at which time the
    pot values are read and saved and the lamp is 
    adjusted accordingly.
  
  TO-DO --------------------------------------------------
  1) Build physical lamp
  2) Read / save tag and color data to flash mem using
     PROGMEM
  
**********************************************************/

//-- External libraries ----------------------------------------------------------
#include <SoftwareSerial.h>

//-- Pin assignments --------------------------------------------------------------
//- RFID reader
#include "WProgram.h"
void setup();
void loop();
void normalMode();
void programmingMode();
int validateTag( char* tag );
void loadPreference( char* tag );
void changeColor( int color[] );
void flash( int color[] );
int rfidEnablePin = 14;  // LOW = enable RFID reader; Analog pin 0
int rfidRxPin     = 15;  // Receives serial data from RFID reader; Analog pin 1
int rfidTxPin     = 16;  // Needed to initialize SoftwareSerial class, not used

//- RGB LED
int redLED   = 9;   // Red channel (output)
int greenLED = 10;  // Green channel (output)
int blueLED  = 11;  // Blue channel (output)

//- Programming module
int programPin = 3;   // HIGH = programming module is attached
int setPin     = 4;   // HIGH = light levels set on prog. module
int redPot     = 17;  // Red channel (input); Analog pin 3
int greenPot   = 18;  // Green channel (input); Analog pin 4
int bluePot    = 19;  // Blue channel (input); Analog pin 5

//-- Objects ---------------------------------------------------------------------
SoftwareSerial rfidSerial = SoftwareSerial(rfidRxPin, rfidTxPin);

//-- RFID tag preferences --------------------------------------------------------
int numTags = 2;
int tagLength = 10;
const int maxTags = 20;

char* tags[maxTags] = {"23920592821","295820233"};
int colors[maxTags][3] = {{255,255,255},
                          {255,0,0}};

//-- Program constants -----------------------------------------------------------
int red[] = {255,0,0};

//-- Flags -----------------------------------------------------------------------
boolean debug = true;
boolean n_sent = false;
boolean p_sent = false;


//---------------------------------------------------------------------//
//-- Set up the pins and variables  -----------------------------------//
//---------------------------------------------------------------------//
void setup() {
    // Ser up serial communication for debugging @ 2400bps
    if(debug) Serial.begin(2400);

  // Set up the RFID reader
  pinMode( rfidRxPin, INPUT );         // Set up pin to receive serial data from RFID reader
  pinMode( rfidEnablePin, OUTPUT );    // Set up the RFID enable pin
  digitalWrite( rfidEnablePin, LOW );  // Activate RFID reader to start
  rfidSerial.begin( 2400 );            // Start serial communication @ 2400bps with reader
  
    if(debug)    Serial.println("RFID reader initialized");
  
  // Set up the RGB LEDs
  pinMode( redLED,   OUTPUT );  // Set up the red channel pin
  pinMode( greenLED, OUTPUT );  // Set up the green channel pin
  pinMode( blueLED,  OUTPUT );  // Set up the blue channel pin
  
    if(debug)    Serial.println("RGB lamp initialized");
    
  // Set up programming mode pins
  pinMode( programPin, INPUT );
  pinMode( setPin,     INPUT );
  pinMode( redPot,     INPUT );
  pinMode( greenPot,   INPUT );
  pinMode( bluePot,    INPUT );
  
}  


void loop() {
  
  // programPin is high only when programming module is attached
  if( digitalRead(programPin) != HIGH )
    normalMode();
  else
    programmingMode();    
  
} 


//---------------------------------------------------------------------//
//-- NORMAL MODE  -----------------------------------------------------//
//---------------------------------------------------------------------//
void normalMode() {
  if(!n_sent && debug) {
    Serial.println();
    Serial.println("NORMAL MODE");
    n_sent = true;
    p_sent = false;
  }
  
  char tag[tagLength];        // Complete RFID tag
  int bytesread, val;  // Number of bytes read from RFID reader  

  // Read in unique non-header bits
  if((val = rfidSerial.read()) == 10) {
    bytesread = 0;
    while(bytesread < tagLength) {
      val = rfidSerial.read();
      if((val == 10)||(val == 13))
        break;
      tag[bytesread] = val;
      bytesread++;
    }
  }

  // Lookup the tag once its been read
  if(bytesread == tagLength) {
    if(debug) {
      Serial.print("Tag found: ");
      for(int i=0; i<tagLength; i++)
        Serial.print(tag[i], BYTE);
      Serial.println();
    }
        
    //loadPreference(tag);
  }
}



//---------------------------------------------------------------------//
//-- PROGRAMMING MODE -------------------------------------------------//
//---------------------------------------------------------------------//
void programmingMode() {
  if(!p_sent && debug) {
    Serial.println();
    Serial.println("PROGRAMMING MODE");
    p_sent = true;
    n_sent = false;
  }  
  
  char tag[10];
  int bytesread, val, index;
  
  // Poll RFID reader for header bit
  if((val = rfidSerial.read()) == 10) {
    bytesread = 0;
    while(bytesread < 10) {         // Read in the next 10 bits
      val = rfidSerial.read();
      if((val == 10)||(val == 13))  // Break loop if header or footer bit is read
        break;
      tag[bytesread] = val;
      bytesread++;
    }
  }
  
  // Check if we have read an entire tag
  if(bytesread == 10) {  
      index = validateTag(tag);
      
      if(index >= 0) {
        // Allow LED levels to be adjusted until switch is prsesed
        while( digitalRead(setPin) == LOW ) {
          analogWrite( redLED,   analogRead(redPot)/4 );
          analogWrite( greenLED, analogRead(greenPot)/4 );
          analogWrite( blueLED,  analogRead(bluePot)/4 );
        }      

        // Add in the new color values
        colors[index][0] = analogRead(redPot)/4;
        colors[index][1] = analogRead(greenPot)/4;
        colors[index][2] = analogRead(bluePot)/4;
      }
      
  }
  
}


//---------------------------------------------------------------------//
//-- HELPER FUNCTIONS -------------------------------------------------//  
//---------------------------------------------------------------------//

/**********************************************
 Validate an RFID tag
 - Checks if the given RFID tag is stored
 - Returns tag array index value if valid
 - Returns -1 if invalid. Will also flash
   LED red.
**********************************************/
int validateTag( char* tag ) {
  boolean index = -1;

  for(int i=0; i<numTags; i++) {
    if(strcmp(tags[i],tag)==0) {
      index = i;
      break;
    }
  }
  
  if(index == -1)  flash(red);

  return index;
}

/**********************************************
 Load preference of an RFID tag
 - Looks up the color value associated with
   the given tag and loads the color
**********************************************/
void loadPreference( char* tag ) {
  int index = validateTag(tag);
  
  if(index >= 0)  changeColor( colors[index] );
}


/********************************************
 Change color of LED
 - Takes in one array of ints, representing
   desired RGB values, then updates LED
*********************************************/
void changeColor( int color[] ) {
  if(debug) {
    Serial.print("Changing color to: {R=");
    Serial.print( color[0] );
    Serial.print("} {G=");
    Serial.print( color[1] );
    Serial.print("} {B=");
    Serial.print( color[2] );
    Serial.println("}");
  }
  
  // Change to new color values
  analogWrite( redLED,   color[0] );  // Update red channel
  analogWrite( greenLED, color[1] );  // Update green channel
  analogWrite( blueLED,  color[2] );  // Update blue channel
}


/***********************************************
  Flash
  - Takes in one array of ints representing
    an RGB value. Saves current color values
    of LED, then changes them to the desired
    values for a short time before returning
    back to original color values
***********************************************/
void flash( int color[] ) {
  
  // Save the current color values
  int prevRed   = analogRead( redLED ) / 4;
  int prevGreen = analogRead( greenLED ) / 4;
  int prevBlue  = analogRead( blueLED  ) / 4;
  
  // Change to new color values
  analogWrite( redLED,   color[0] );
  analogWrite( greenLED, color[1] );
  analogWrite( blueLED,  color[2] );
  
  delay( 500 );   // Wait for some time
  
  // Return to previous color values
  analogWrite( redLED,   prevRed   );
  analogWrite( greenLED, prevGreen );
  analogWrite( blueLED,  prevBlue  );
  
}


int main(void)
{
	init();

	setup();
    
	for (;;)
		loop();
        
	return 0;
}

